﻿using System;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace RTTrick.Internals
{
    /// <summary>
    /// 메서드 테이블
    /// </summary>
    [StructLayout(LayoutKind.Sequential)]
    internal unsafe struct MethodTable
    {
        public static unsafe MethodTable* CreateFromType<T>()
        {
            return CreateFromType(typeof(T));
        }
        public static unsafe MethodTable* CreateFromType(Type t)
        {
            if (t == null)
                throw new ArgumentNullException(nameof(t));

            return (MethodTable*)t.TypeHandle.Value;
        }

        #region Inner Types
        // Enum types has imported from MethodTable.h
        [Flags]
        private enum FlagsEnum : UInt32
        {
            UNUSED_ComponentSize_1 = 0x00000001,

            StaticsMask = 0x00000006,
            StaticsMask_NonDynamic = 0x00000000,
            StaticsMask_Dynamic = 0x00000002,   // dynamic statics (EnC, reflection.emit)
            StaticsMask_Generics = 0x00000004,   // generics statics
            StaticsMask_CrossModuleGenerics = 0x00000006, // cross module generics statics (NGen)
            StaticsMask_IfGenericsThenCrossModule = 0x00000002, // helper constant to get rid of unnecessary check

            NotInPZM = 0x00000008,   // True if this type is not in its PreferredZapModule

            GenericsMask = 0x00000030,
            GenericsMask_NonGeneric = 0x00000000,   // no instantiation
            GenericsMask_GenericInst = 0x00000010,   // regular instantiation, e.g. List<String>
            GenericsMask_SharedInst = 0x00000020,   // shared instantiation, e.g. List<__Canon> or List<MyValueType<__Canon>>
            GenericsMask_TypicalInst = 0x00000030,   // the type instantiated at its formal parameters, e.g. List<T>

            HasVariance = 0x00000100,   // This is an instantiated type some of whose type parameters are co- or contra-variant

            HasDefaultCtor = 0x00000200,
            HasPreciseInitCctors = 0x00000400,   // Do we need to run class constructors at allocation time? (Not perf important, could be moved to EEClass

            IsHFA = 0x00000800,   // This type is an HFA (Homogenous Floating-point Aggregate)
            IsRegStructPassed = 0x00000800,   // This type is a System V register passed struct.

            IsByRefLike = 0x00001000,

            // In a perfect world we would fill these flags using other flags that we already have
            // which have a constant value for something which has a component size.
            UNUSED_ComponentSize_5 = 0x00002000,
            UNUSED_ComponentSize_6 = 0x00004000,
            UNUSED_ComponentSize_7 = 0x00008000,

            // As you change the flags in WFLAGS_LOW_ENUM you also need to change this
            // to be up to date to reflect the default values of those flags for the
            // case where this MethodTable is for a String or Array
            //StringArrayValues = SET_TRUE(StaticsMask_NonDynamic) |
            //                              SET_FALSE(NotInPZM) |
            //                              SET_TRUE(GenericsMask_NonGeneric) |
            //                              SET_FALSE(HasVariance) |
            //                              SET_FALSE(HasDefaultCtor) |
            //                              SET_FALSE(HasPreciseInitCctors),

            StringArrayValues = StaticsMask_NonDynamic |
                                          GenericsMask_NonGeneric,

            // The following bits describe mutually exclusive locations of the type
            // in the type hiearchy.
            Category_Mask = 0x000F0000,

            Category_Class = 0x00000000,
            Category_Unused_1 = 0x00010000,
            Category_Unused_2 = 0x00020000,
            Category_Unused_3 = 0x00030000,

            Category_ValueType = 0x00040000,
            Category_ValueType_Mask = 0x000C0000,
            Category_Nullable = 0x00050000, // sub-category of ValueType
            Category_PrimitiveValueType = 0x00060000, // sub-category of ValueType, Enum or primitive value type
            Category_TruePrimitive = 0x00070000, // sub-category of ValueType, Primitive (ELEMENT_TYPE_I, etc.)

            Category_Array = 0x00080000,
            Category_Array_Mask = 0x000C0000,
            // Category_IfArrayThenUnused                 = 0x00010000, // sub-category of Array
            Category_IfArrayThenSzArray = 0x00020000, // sub-category of Array

            Category_Interface = 0x000C0000,
            Category_Unused_4 = 0x000D0000,
            Category_Unused_5 = 0x000E0000,
            Category_Unused_6 = 0x000F0000,

            Category_ElementTypeMask = 0x000E0000, // bits that matter for element type mask


            HasFinalizer = 0x00100000, // instances require finalization

            IDynamicInterfaceCastable = 0x00200000, // class implements IDynamicInterfaceCastable interface

            ICastable = 0x00400000, // class implements ICastable interface

            HasIndirectParent = 0x00800000, // m_pParentMethodTable has double indirection

            ContainsPointers = 0x01000000,

            HasTypeEquivalence = 0x02000000, // can be equivalent to another type

            IsTrackedReferenceWithFinalizer = 0x04000000,

            HasCriticalFinalizer = 0x08000000, // finalizer must be run on Appdomain Unload
            Collectible = 0x10000000,
            ContainsGenericVariables = 0x20000000,   // we cache this flag to help detect these efficiently and
                                                     // to detect this condition when restoring

            ComObject = 0x40000000, // class is a com object

            HasComponentSize = 0x80000000,   // This is set if component size is used for flags.

            // Types that require non-trivial interface cast have this bit set in the category
            NonTrivialInterfaceCast = Category_Array
                                                 | ComObject
                                                 | ICastable
                                                 | IDynamicInterfaceCastable
                                                 | Category_ValueType
        }

        // Enum types has imported from MethodTable.h
        [Flags]
        private enum Flags2Enum : UInt16
        {
            // The following bits describe usage of optional slots. They have to stay
            // together because of we index using them into offset arrays.
            MultipurposeSlotsMask = 0x001F,
            HasPerInstInfo = 0x0001,
            HasInterfaceMap = 0x0002,
            HasDispatchMapSlot = 0x0004,
            HasNonVirtualSlots = 0x0008,
            HasModuleOverride = 0x0010,

            IsZapped = 0x0020, // This could be fetched from m_pLoaderModule if we run out of flags

            IsPreRestored = 0x0040, // Class does not need restore
                                    // This flag is set only for NGENed classes (IsZapped is true)

            HasModuleDependencies = 0x0080,

            IsIntrinsicType = 0x0100,

            RequiresDispatchTokenFat = 0x0200,

            HasCctor = 0x0400,
            HasVirtualStaticMethods = 0x0800,

            RequiresAlign8 = 0x1000, // Type requires 8-byte alignment (only set on platforms that require this and don't get it implicitly)

            HasBoxedRegularStatics = 0x2000, // GetNumBoxedRegularStatics() != 0

            HasSingleNonVirtualSlot = 0x4000,

            DependsOnEquivalentOrForwardedStructs = 0x8000, // Declares methods that have type equivalent or type forwarded structures in their signature
        }
        #endregion

        #region Private Fields
        private FlagsEnum _flags;
        private Int32 _baseSize;
        private Flags2Enum _flags2;
        private Int16 _token;

        private Int16 _numVirtuals;
        private Int16 _numInterfaces;

        private MethodTable* _parentMethodTable;

        private IntPtr _loaderModule;
        private IntPtr _writableData;

        [StructLayout(LayoutKind.Explicit)]
        private struct Union1
        {
            [FieldOffset(0)]
            private IntPtr _eeClass;
            [FieldOffset(0)]
            private IntPtr _canonMT;
        }

        [DebuggerBrowsable(DebuggerBrowsableState.Never)]
        private Union1 _union1;

        [StructLayout(LayoutKind.Explicit)]
        private struct Union2
        {
            [FieldOffset(0)]
            private IntPtr _perInstInfo;
            [FieldOffset(0)]
            private IntPtr _elementTypeHnd;
            [FieldOffset(0)]
            private IntPtr _multiPurposeSlot1;
        }
        [DebuggerBrowsable(DebuggerBrowsableState.Never)]
        private Union2 _union2;

        [StructLayout(LayoutKind.Explicit)]
        private struct Union3
        {
            [FieldOffset(0)]
            private IntPtr _pInterfaceMap;
            [FieldOffset(0)]
            private IntPtr _multiPurposeSlot2;
        }
        [DebuggerBrowsable(DebuggerBrowsableState.Never)]
        private Union3 _union3;
        #endregion

        public bool IsGenericType()
        {
            throw new NotImplementedException();
        }

        public bool IsArrayType()
        {
            throw new NotImplementedException();
        }

        public int GetInterfaceNumber()
        {
            throw new NotImplementedException();
        }

        public int GetObjectSize()
        {
            throw new NotImplementedException();
        }
    }
}
